"""
Exception hierarchy for pynng.  The base of the hierarchy is NNGException.
Every exception has a corresponding errno attribute which can be checked.

Generally, each number in nng_errno_enum corresponds with an Exception type.


"""

from ._nng import ffi, lib as nng


class NNGException(Exception):
    """The base exception for any exceptional condition in the nng bindings."""
    def __init__(self, msg, errno):
        super().__init__(msg)
        self.errno = errno


class Interrupted(NNGException):  # NNG_EINTR
    pass


class NoMemory(NNGException):  # NNG_ENOMEM
    pass


class InvalidOperation(NNGException):  # NNG_EINVAL
    pass


class Busy(NNGException):  # NNG_EBUSY
    pass


class Timeout(NNGException):  # NNG_ETIMEDOUT
    pass


class ConnectionRefused(NNGException):  # NNG_ECONNREFUSED
    pass


class Closed(NNGException):  # NNG_ECLOSED
    pass


class TryAgain(NNGException):  # NNG_EAGAIN
    pass


class NotSupported(NNGException):  # NNG_ENOTSUP
    pass


class AddressInUse(NNGException):  # NNG_EADDRINUSE
    pass


class BadState(NNGException):  # NNG_ESTATE
    pass


class NoSuchFile(NNGException):  # NNG_ENOENT
    pass


class ProtocolError(NNGException):  # NNG_EPROTO
    pass


class DestinationUnreachable(NNGException):  # NNG_EUNREACHABLE
    pass


class AddressInvalid(NNGException):  # NNG_EADDRINVAL
    pass


class PermissionDenied(NNGException):  # NNG_EPERM
    pass


class MessageTooLarge(NNGException):  # NNG_EMSGSiZE
    pass


class ConnectionReset(NNGException):  # NNG_ECONNRESET
    pass


class ConnectionAborted(NNGException):  # NNG_ECONNABORTED
    pass


class Canceled(NNGException):  # NNG_ECANCELED
    pass


class OutOfFiles(NNGException):  # NNG_ENOFILES
    pass


class OutOfSpace(NNGException):  # NNG_ENOSPC
    pass


class AlreadyExists(NNGException):  # NNG_EEXIST
    pass


class ReadOnly(NNGException):  # NNG_EREADONLY
    pass


class WriteOnly(NNGException):  # NNG_EWRITEONLY
    pass


class CryptoError(NNGException):  # NNG_ECRYPTO
    pass


class AuthenticationError(NNGException):  # NNG_EPEERAUTH
    pass


class NoArgument(NNGException):  # NNG_ENOARG
    pass


class Ambiguous(NNGException):  # NNG_EAMBIGUOUS
    pass


class BadType(NNGException):  # NNG_EBADTYPE
    pass


class Internal(NNGException):  # NNG_EWRITEONLY
    pass


# maps exceptions from the enum nng_errno_enum to Exception classes
EXCEPTION_MAP = {
    nng.NNG_EINTR: Interrupted,
    nng.NNG_ENOMEM: NoMemory,
    nng.NNG_EINVAL: InvalidOperation,
    nng.NNG_EBUSY: Busy,
    nng.NNG_ETIMEDOUT: Timeout,
    nng.NNG_ECONNREFUSED: ConnectionRefused,
    nng.NNG_ECLOSED: Closed,
    nng.NNG_EAGAIN: TryAgain,
    nng.NNG_ENOTSUP: NotSupported,
    nng.NNG_EADDRINUSE: AddressInUse,
    nng.NNG_ESTATE: BadState,
    nng.NNG_ENOENT: NoSuchFile,
    nng.NNG_EPROTO: ProtocolError,
    nng.NNG_EUNREACHABLE: DestinationUnreachable,
    nng.NNG_EADDRINVAL: AddressInvalid,
    nng.NNG_EPERM: PermissionDenied,
    nng.NNG_EMSGSIZE: MessageTooLarge,
    nng.NNG_ECONNRESET: ConnectionReset,
    nng.NNG_ECONNABORTED: ConnectionAborted,
    nng.NNG_ECANCELED: Canceled,
    nng.NNG_ENOFILES: OutOfFiles,
    nng.NNG_ENOSPC: OutOfSpace,
    nng.NNG_EEXIST: AlreadyExists,
    nng.NNG_EREADONLY: ReadOnly,
    nng.NNG_EWRITEONLY: WriteOnly,
    nng.NNG_ECRYPTO: CryptoError,
    nng.NNG_EPEERAUTH: AuthenticationError,
    nng.NNG_ENOARG: NoArgument,
    nng.NNG_EAMBIGUOUS: Ambiguous,
    nng.NNG_EBADTYPE: BadType,
    nng.NNG_EINTERNAL: Internal,
}


def check_err(err):
    """
    Raises an exception if the return value of an nng_function is nonzero.

    The enum nng_errno_enum is defined in nng.h

    """
    # fast path for success
    if not err:
        return

    msg = nng.nng_strerror(err)
    string = ffi.string(msg)
    string = string.decode()
    exc = EXCEPTION_MAP.get(err, NNGException)
    raise exc(string, err)





import os
import shutil
import subprocess
import sys

import setuptools.command.build_py


THIS_DIR = os.path.dirname(__file__)


def build_nng_lib():
    # cannot import build_pynng at the top level becuase cffi may not be
    # installed yet (since it is a dependency, and this script installs
    # dependencies).  Bootstrapping!
    import build_pynng
    if os.path.exists(build_pynng.objects[0]):
        # the object file we were planning on building already exists; we'll
        # just use it!
        return
    if sys.platform == 'win32':
        is_64bit = sys.maxsize > 2**32
        major, minor, *_ = sys.version_info
        build_nng_script = os.path.join(THIS_DIR, 'build_nng.bat')
        needs_shell = True
        # pick the correct cmake generator, based on the Python version.
        # from https://wiki.python.org/moin/WindowsCompilers for Python
        # version, and cmake --help for list of CMake generator names

        # If ninja build system is installed, use it, since it's way faster
        # (this is especially important when feeling impatient for CI builds)
        if shutil.which('ninja') and (major, minor) in ((3, 5), (3, 6), (3, 7)):
            # gotta soruce the correct vcvarsall!
            which_vcvars = {
                (3, 5): r'Microsoft Visual Studio 14.0\VC',
                (3, 6): r'Microsoft Visual Studio 14.0\VC',
                (3, 7): r'Microsoft Visual Studio 14.0\VC',
            }
            vcvarsall = os.path.join(
                r'C:\Program Files (x86)',
                which_vcvars[(major, minor)],
                'vcvarsall.bat',
            )
            if is_64bit:
                gen = 'amd64'
            else:
                gen = 'x86'
            cmd = '"{}" {} && {} Ninja'.format(vcvarsall, gen, build_nng_script)
            print('-------------------')
            print(cmd)
            print('-------------------')

        else:
            cmake_generators = {
                (3, 5): 'Visual Studio 14 2015',
                (3, 6): 'Visual Studio 14 2015',
                (3, 7): 'Visual Studio 14 2015',
            }
            gen = cmake_generators[(major, minor)]

            if is_64bit:
                gen += ' Win64'

            cmd = [build_nng_script, gen]

    else:
        # on Linux, build_nng.sh selects ninja if available
        script = os.path.join(THIS_DIR, 'build_nng.sh')
        cmd = ['/bin/bash', script]
        needs_shell = False

    # shell=True is required for Windows
    subprocess.check_call(cmd, shell=needs_shell)


# TODO: this is basically a hack to get something to run before running cffi
# extnsion builder. subclassing something else would be better!
class BuildCommand(setuptools.command.build_py.build_py):
    """Custom build command."""

    def run(self):
        build_nng_lib()
        super(BuildCommand, self).run()


with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    cmdclass={
        'build_py': BuildCommand,
    },
    name='pynng',
    version='0.1.1',
    author='Cody Piersall',
    author_email='cody.piersall@gmail.com',
    description='Networking made simply using nng',
    long_description=long_description,
    license='MIT',
    keywords='networking nng nanomsg zmq messaging message',
    long_description_content_type='text/markdown',
    url='https://github.com/codypiersall/pynng',
    packages=setuptools.find_packages(),
    classifiers=(
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Networking',
    ),
    setup_requires=['cffi'],
    install_requires=['cffi'],
    cffi_modules=['build_pynng.py:ffibuilder'],
)

